#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <string.h>

#include "wlioctl.h"

/* What should be in wlapi.h */
#define FALSE   0
#define TRUE    1

typedef unsigned long           DWORD;
typedef unsigned short          WORD;
typedef unsigned char           BYTE;
typedef int                     BOOL;

/* */
/* define CCR in EXTRA MIB */
/* */
#define CCR_AP          0
#define CCR_INFRA       1
#define CCR_ADHOC       2
#define CCR_IAP         3

/* Ethernet MAC Address, BSS_ID, or ESS_ID */
/* With this, we may simply write "x=y;" instead of "memcpy(x, y, 6);" */
/* It's more efficiency with compiler's optimization and more clearly  */
typedef struct _MADDR {
    BYTE    b0;
    BYTE    b1;
    BYTE    b2;
    BYTE    b3;
    BYTE    b4;
    BYTE    b5;
} MADDR;

typedef struct _WL_Monitor {
    MADDR   MAddr;
    WORD    RSSI;
    DWORD   TimeStamp;
} WL_Monitor, *PWL_Monitor;
   

int             g_fd;
struct ifreq    g_ifreq;
WL_IOCTL_BLK    *g_pBlk;
unsigned char   g_userdata[1024];

typedef struct _WL_CMD {
    const char *cpszCmd;
    int         nQueryArgc;
    int         nSetArgc;
    BOOL        (*CallFunc)();
} WL_CMD;

BOOL WL_CmdReset(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery || strcasecmp(cpszArgv[0], "now") != 0) {
        printf("ERROR: RESET is write-only! Try 'reset now'!\n");
        return FALSE;
    }

    g_pBlk->uCmd = WL_IOCTL_SET_RESET;
    *g_pBlk->pData = 0;
    
    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("RESET NOW\n");
    return TRUE;
}

BOOL WL_CmdFlash(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery || strcasecmp(cpszArgv[0], "write") != 0) {
        printf("ERROR: FRASH is write-only! Try 'flash write'!\n");
        return FALSE;
    }

    g_pBlk->uCmd = WL_IOCTL_SET_FLASH;
    *g_pBlk->pData = 0;
    
    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("FLASH WRITE\n");
    return TRUE;
}

BOOL WL_CmdVersion(BOOL bQuery, const char *cpszArgv[])
{
    if (!bQuery) {
        printf("ERROR: 'VERSION' is read-only!\n");
        return FALSE;
    }
        
    g_pBlk->uCmd = WL_IOCTL_GET_VERSION;
    *g_pBlk->pData = 0;
    
    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("VERSION = %s\n", (const char *) g_pBlk->pData);
    return TRUE;
}

BOOL WL_CmdDebug(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery) {
        /* Get Debug Flag */    
        g_pBlk->uCmd = WL_IOCTL_GET_DBGFLAG;
    }
    else {
        unsigned long ulDbgFlag;
        
        sscanf(cpszArgv[0], "%lx", &ulDbgFlag);
        printf("Set DEBUG = 0x%lx\n", ulDbgFlag);
        
        g_pBlk->uCmd = WL_IOCTL_SET_DBGFLAG;
        g_pBlk->uLen = sizeof(ulDbgFlag);
        memcpy(g_pBlk->pData, &ulDbgFlag, sizeof(ulDbgFlag));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("DEBUG = 0x%lx\n", *((unsigned long *) g_pBlk->pData));
    return TRUE;
}

BOOL WL_CmdChannel(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery) {
        /* Get Debug Flag */    
        g_pBlk->uCmd = WL_IOCTL_GET_CHANNEL;
    }
    else {
        long lChannel;
        
        sscanf(cpszArgv[0], "%ld", &lChannel);
        printf("Set CHANNEL = %ld\n", lChannel);
        
        if (lChannel < 1 || lChannel > 12) {
            printf("ERROR: CHANNEL must be between 1 and 12\n");
            return FALSE;
        } 
        g_pBlk->uCmd = WL_IOCTL_SET_CHANNEL;
        g_pBlk->uLen = sizeof(lChannel);
        memcpy(g_pBlk->pData, &lChannel, sizeof(lChannel));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("CHANNEL = %ld\n", *((unsigned long *) g_pBlk->pData));
    return TRUE;
}


BOOL WL_CmdSpeed(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_SPEED;
    }
    else {
        long lSpeed;
        
        sscanf(cpszArgv[0], "%ld", &lSpeed);
        printf("Set SPEED = %ld\n", lSpeed);
        
        if (lSpeed != 1 && lSpeed != 2) {
            printf("ERROR: SPEED only accepts '1' for 1 Mbps or '2' for 2 Mbps!\n");
            return FALSE;
        } 
        g_pBlk->uCmd = WL_IOCTL_SET_SPEED;
        g_pBlk->uLen = sizeof(lSpeed);
        memcpy(g_pBlk->pData, &lSpeed, sizeof(lSpeed));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("SPEED = %ld\n", *((unsigned long *) g_pBlk->pData));
    return TRUE;
}


BOOL WL_CmdMode(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_MODE;
    }
    else {
        long lMode;
        
        sscanf(cpszArgv[0], "%ld", &lMode);
        printf("Set MODE = %ld\n", lMode);
        
        if (lMode < CCR_AP || lMode > CCR_IAP) {
            printf("ERROR: mode %ld is unknown!\n", lMode);
            return FALSE;
        }
        
        g_pBlk->uCmd = WL_IOCTL_SET_MODE;
        g_pBlk->uLen = sizeof(lMode);
        memcpy(g_pBlk->pData, &lMode, sizeof(lMode));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("MODE = %ld\n", *((unsigned long *) g_pBlk->pData));
    return TRUE;
}

BOOL WL_CmdRoaming(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_ROAMING;
    }
    else {
        long lRoaming;
        
        sscanf(cpszArgv[0], "%ld", &lRoaming);
        printf("Set ROAMING = %ld\n", lRoaming);
        
        if (lRoaming != 0 && lRoaming != 1) {
            printf("ERROR: ROAMING only accepts '0' for off or '1' for on!\n");
            return FALSE;
        } 
        g_pBlk->uCmd = WL_IOCTL_SET_ROAMING;
        g_pBlk->uLen = sizeof(lRoaming);
        memcpy(g_pBlk->pData, &lRoaming, sizeof(lRoaming));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("ROAMING = %ld\n", *((unsigned long *) g_pBlk->pData));
    return TRUE;
}


BOOL WL_CmdBSSID(BOOL bQuery, const char *cpszArgv[])
{
    unsigned char *pID;
    
    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_BSSID;
    }
    else {
        unsigned char BSSID[6];
        int i;
        
        for (i = 0; i < sizeof(BSSID); i++) {
            int nHex;
            sscanf(cpszArgv[i], "%x", &nHex);

            if (nHex < 0 || nHex > 255) {
                printf("ERROR: BSSID must be 6 sets of 00-FF\n");
                return FALSE;
            }
            
            BSSID[i] = nHex;
        }
            
        printf("Set BSSID = %02x %02x %02x %02x %02x %02x\n", 
                BSSID[0], BSSID[1], BSSID[2], BSSID[3], BSSID[4], BSSID[5]);
        
        g_pBlk->uCmd = WL_IOCTL_SET_BSSID;
        g_pBlk->uLen = sizeof(BSSID);
        memcpy(g_pBlk->pData, BSSID, sizeof(BSSID));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    pID = g_pBlk->pData;
    printf("BSSID = %02x %02x %02x %02x %02x %02x\n", 
            pID[0], pID[1], pID[2], pID[3], pID[4], pID[5]);
    return TRUE;
}

BOOL WL_CmdESSID(BOOL bQuery, const char *cpszArgv[])
{
    /* ESSID is only 6 bytes, but user input may exceed */
    unsigned char ESSID[40];

    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_ESSID;
    }
    else {
        long lESSID;
        
        sscanf(cpszArgv[0], "%ld", &lESSID);
        
        if (lESSID < 0 || lESSID > 999999) {
            printf("ERROR: ESSID must be between 000000 and 999999\n");
            return FALSE;
        }
        
        sprintf(ESSID, "%06ld", lESSID);
        printf("Set ESSID = %s\n", ESSID);
        
        g_pBlk->uCmd = WL_IOCTL_SET_ESSID;
        g_pBlk->uLen = 6;
        memcpy(g_pBlk->pData, &ESSID, g_pBlk->uLen);
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }

    /* valid data is first 6 bytes ( 0 - 5 ) */
    memcpy(ESSID, g_pBlk->pData, 6);
    ESSID[6] = 0;   
    printf("ESSID = %s\n", ESSID);
    return TRUE;
}

BOOL WL_CmdSecurity(BOOL bQuery, const char *cpszArgv[])
{
    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_SECURITY;
    }
    else {
        long lSecurity;
        
        sscanf(cpszArgv[0], "%ld", &lSecurity);
        printf("Set SECURITY = %06ld\n", lSecurity);
        
        if ((lSecurity < 0 || lSecurity > 999999) && lSecurity != -1) {
            printf("ERROR: SECURITY must be between 000000 and 999999, or -1 for off\n");
            return FALSE;
        }
        
        g_pBlk->uCmd = WL_IOCTL_SET_SECURITY;
        g_pBlk->uLen = sizeof(lSecurity);
        memcpy(g_pBlk->pData, &lSecurity, sizeof(lSecurity));
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("SECURITY = %06ld\n", *((unsigned long *) g_pBlk->pData));
    return TRUE;
}

BOOL WL_CmdMonitor(BOOL bQuery, const char *cpszArgv[])
{
    WL_Monitor *pMonitor = (WL_Monitor *) g_pBlk->pData;
    
    if (bQuery) {
        g_pBlk->uCmd = WL_IOCTL_GET_MONITOR;
    }
    else {
        int i;
        
        for (i = 0; i < sizeof(MADDR); i++) {
            BYTE *pByte;
            int nHex;
            sscanf(cpszArgv[i], "%x", &nHex);

            if (nHex < 0 || nHex > 255) {
                printf("ERROR: MONITOR Address must be 6 sets of 00-FF\n");
                return FALSE;
            }
            
            pByte = (BYTE *) &pMonitor->MAddr;
            pByte[i] = nHex;
        }
            
        printf("Set MONITOR ADDRESS = %02x %02x %02x %02x %02x %02x\n", 
                pMonitor->MAddr.b0, pMonitor->MAddr.b1, pMonitor->MAddr.b2, 
                pMonitor->MAddr.b3, pMonitor->MAddr.b4, pMonitor->MAddr.b5);
        
        g_pBlk->uCmd = WL_IOCTL_SET_MONITOR;
        g_pBlk->uLen = sizeof(pMonitor->MAddr);
    }

    if (ioctl(g_fd, SIOC_WL_IOCTL, &g_ifreq) == -1) {
        perror("ERROR");
        return FALSE;
    }
    
    printf("MONITOR ADDRESS = %02x %02x %02x %02x %02x %02x\n", 
            pMonitor->MAddr.b0, pMonitor->MAddr.b1, pMonitor->MAddr.b2, 
            pMonitor->MAddr.b3, pMonitor->MAddr.b4, pMonitor->MAddr.b5);

    if (bQuery)
        printf("MONITOR RSSI = %u\nMONITOR TIMESTAMP = %lu0 ms AGO\n", 
                pMonitor->RSSI, pMonitor->TimeStamp);
   
    return TRUE;
}

WL_CMD g_WlCmd[] = {
    { "version",    1, 0, WL_CmdVersion },  /* query / (read only)          */
    { "reset",      1, 1, WL_CmdReset},     /* (write only) / now           */
    { "flash",      1, 1, WL_CmdFlash},     /* (write only) / now           */
    { "debug",      1, 1, WL_CmdDebug },    /* query / 0-FFFFFFFF           */
    { "channel",    1, 1, WL_CmdChannel },  /* query / 1-12                 */
    { "speed",      1, 1, WL_CmdSpeed },    /* query / 1-2                  */
    { "mode",       1, 1, WL_CmdMode },     /* query / 1 (adhoc) / 2 (infra)*/
    { "roaming",    1, 1, WL_CmdRoaming },  /* query / 0 (off) / 1 (on)     */
    { "BSSID",      1, 6, WL_CmdBSSID },    /* query / 00 00 00 00 00 00-FF FF FF FF FF FF */
    { "ESSID",      1, 1, WL_CmdESSID },    /* query / 000000-999999        */
    { "security",   1, 1, WL_CmdSecurity }, /* query / 000000-999999(on) / -1 (off) */
    { "monitor",    1, 6, WL_CmdMonitor },  /* query / 00 00 00 00 00 00-FF FF FF FF FF FF */
    { NULL, 0, 0, NULL }
};

int FindCommandIndex(const char *cpszInput)
{
    int         i;
    const char *cpszCmd;
    
    for (i = 0; (cpszCmd = g_WlCmd[i].cpszCmd) != NULL; i++) {
        if (strcasecmp(cpszInput, cpszCmd) == 0)
            return i;
    }
    
    return -1;
}

int main(int argc, char *argv[])
{
    int i;
    
    if (getuid() != 0) {
        printf("\nYou must be root to run %s\n\n", argv[0]);
        return 1;
    }
    
    if (argc < 4) {
        printf("\nUsage: %s interface cmd1 argv11 [argv12, ...] [cmd2 argv21 [argv22, ...]]\n\n", argv[0]);
        return 2;
    }
    
    if ((g_fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        perror("ERROR - socket()");
        return 3;
    }

    strcpy(g_ifreq.ifr_name, argv[1]);
    g_pBlk = (WL_IOCTL_BLK *) &g_ifreq.ifr_data;
    g_pBlk->pData = g_userdata;

    /* program & interface had been processed */
    for (i = 2; i < argc; ) {
        int nNeededArgc;
        int bQuery;
        BOOL bResult;
        int nCmdIndex = FindCommandIndex(argv[i]);
        
        if (nCmdIndex == -1) {
            printf("Unknow command - %s\n", argv[i]);
            return -1;
        }
        
        i++;
        if (i >= argc) {
            printf("Insufficient arguments for command %s (argv %d)\n", 
                    g_WlCmd[nCmdIndex].cpszCmd, i-1);
            return -1;
        }
        
        if (strcasecmp(argv[i], "query") == 0) {
            bQuery = TRUE;
            nNeededArgc = g_WlCmd[nCmdIndex].nQueryArgc;
        }
        else {
            bQuery = FALSE;
            nNeededArgc = g_WlCmd[nCmdIndex].nSetArgc;
            if (nNeededArgc == 0) {
                printf("Command '%s' is read only\n", g_WlCmd[nCmdIndex].cpszCmd);
                return -1;
            } 
        }
        
        if (argc - i < nNeededArgc) {
            printf("Insufficient arguments for command '%s' (at least %d)\n",
                    g_WlCmd[nCmdIndex].cpszCmd, nNeededArgc);
            return -1;
        }
            
        bResult = (*g_WlCmd[nCmdIndex].CallFunc)(bQuery, &argv[i]);
        if (!bResult) {
            printf("Command '%s' failed!!\n\n", 
                    g_WlCmd[nCmdIndex].cpszCmd);
            return -1;
        }
        
        i += nNeededArgc;
    }
        
    return 0;
}


